/*
 * Copyright (c) 2007, Jim C. Nasby (decibel@decibel.org)
 * All rights reserved.
 *
 * $Id: functions.sql 37 2008-11-05 20:19:20Z decibel $
 */

CREATE OR REPLACE FUNCTION gather.table_list() RETURNS SETOF text LANGUAGE plpgsql AS $table_list$
BEGIN
    -- We'll be assuming the same timestamp for stat and statio entries for tables, so do them together
    RETURN NEXT 'pg_stat_user_tables';
    RETURN NEXT 'pg_statio_user_tables';

    -- Same for indexes
    RETURN NEXT 'pg_stat_user_indexes';
    RETURN NEXT 'pg_statio_user_indexes';

    RETURN NEXT 'pg_statio_user_sequences';
END;
$table_list$;

CREATE OR REPLACE FUNCTION gather.all(
    p_update_rrs boolean
) RETURNS void LANGUAGE plpgsql SECURITY DEFINER AS $gather_stats$
DECLARE
    v_cluster       public.clusters;
    v_table_name    record;
BEGIN
    -- Create all the temp tables
    FOR v_table_name IN SELECT * FROM gather.table_list()
    LOOP
        PERFORM gather.create_table( v_table_name.table_list );
    END LOOP;

    FOR v_cluster IN SELECT * FROM public.clusters
    LOOP
        PERFORM gather.cluster( v_cluster );
    END LOOP;

    -- Summarize all the stats and insert them into the final tables
    PERFORM gather.integrate_stats();

    -- Update RRS
    IF p_update_rrs THEN
        PERFORM rrs.update();
    END IF;
END;
$gather_stats$;
CREATE OR REPLACE FUNCTION gather.all() RETURNS void LANGUAGE sql AS 'SELECT gather.all( true )';


CREATE OR REPLACE FUNCTION gather.cluster(
    p_cluster_name  public.clusters.cluster_name%TYPE
) RETURNS void LANGUAGE plpgsql SECURITY DEFINER AS $gather_stats$
DECLARE
    v_cluster       public.clusters;
    v_table_name    record;
BEGIN
    FOR v_cluster IN SELECT * FROM public.clusters WHERE cluster_name = p_cluster_name
    LOOP
        PERFORM gather.cluster( v_cluster );
    END LOOP;

    -- Summarize all the stats and insert them into the final tables
    PERFORM gather.integrate_stats();
END;
$gather_stats$;
CREATE OR REPLACE FUNCTION gather.cluster(
    p_cluster_id  public.clusters.cluster_id%TYPE
) RETURNS void LANGUAGE plpgsql SECURITY DEFINER AS $gather_stats$
DECLARE
    v_cluster       public.clusters;
    v_table_name    record;
BEGIN
    FOR v_cluster IN SELECT * FROM public.clusters WHERE cluster_id = p_cluster_id
    LOOP
        PERFORM gather.cluster( v_cluster );
    END LOOP;

    -- Summarize all the stats and insert them into the final tables
    PERFORM gather.integrate_stats();
END;
$gather_stats$;

CREATE OR REPLACE FUNCTION gather.cluster(
    p_cluster public.clusters
) RETURNS void LANGUAGE plpgsql AS $gather_cluster$
DECLARE
    v_database          raw.databases;
    v_return_types      text;
    sql                 text;
    rsql                text;
    v_idle_stats_count  bigint;
BEGIN
    -- Connect to postgres database in cluster
    IF NOT tools.db_connect( p_cluster, 'postgres' ) THEN
        RETURN;
    END IF;

    -- Create / truncate our temp table
    PERFORM gather.create_table( 'cluster_databases' );

    -- See if our idle_stats function exists
    rsql := $$SELECT count(1)
                    FROM pg_catalog.pg_proc
                    WHERE proname = 'idle_stats'
                        AND pg_catalog.format_type(prorettype, NULL) = 'record'
                        AND pronargs=0
                        AND NOT proisagg$$
    ;
    RAISE DEBUG 'rsql: %', rsql;
    SELECT INTO v_idle_stats_count count
        FROM dblink( rsql ) AS ( count bigint )
    ;

    /* 
     * Grab pg_stat_database, use it's contents to drive calling gather.database. It's critical that
     * this SELECT doesn't reference anything in pg_database (unless we handle
     * the return type stuff ourselves), because table_return_type() only
     * handles single tables.
     */
    rsql := $$SELECT idle, idle_in_transaction, s.*
        FROM pg_database d JOIN pg_stat_database s ON ( d.oid = s.datid )
            LEFT JOIN ( $$
    ;
    
    -- If we can, use our custom stats function to pull idle information
    IF v_idle_stats_count > 0 THEN
        -- If we found more than one function, throw a warning, but go ahead and use it
        IF v_idle_stats_count > 1 THEN
            RAISE WARNING 'More than one idle_stats function found in cluster %', p_cluster.cluster_name;
        END IF;
        rsql := rsql || $$SELECT * FROM idle_stats() AS idle_stats( datid oid, idle int, idle_in_transaction int )$$;
    ELSE
        rsql := rsql || $$SELECT datid, sum(idle) AS idle, sum(idle_in_transaction) AS idle_in_transaction
                    FROM (SELECT datid
                                , CASE WHEN current_query='<IDLE>' THEN 1 ELSE 0 END AS idle
                                , CASE WHEN current_query='<IDLE> in transaction' THEN 1 ELSE 0 END AS idle_in_transaction
                            FROM pg_stat_activity
                        ) a
                    WHERE NOT EXISTS( SELECT *
                                            FROM pg_stat_activity
                                            WHERE current_query IN ( '<command string not enabled>', '<insufficient privilege>' ) )
                    GROUP BY datid$$
        ;
    END IF;
    rsql := rsql || $$
                ) i USING ( datid )
        WHERE d.datallowconn AND NOT d.datistemplate AND d.datname NOT IN ( '$$
            || array_to_string(array_prepend('postgres', coalesce(p_cluster.exclude_databases, '{}')), $$', '$$) || $$' )$$
    ;
    RAISE DEBUG 'rsql: %', rsql;

    -- Figure out what data will be coming back. This only takes single tables into account.
    v_return_types := 'idle int, idle_in_transaction int, ' || tools.table_return_type( 'pg_stat_database' );

    -- Load the data into our temp table. See the comment in gather.table for why we're using timeofday() instead of now().
    sql := $x$INSERT INTO cluster_databases_t SELECT '$x$ || timeofday() || $x$'::timestamptz, * FROM dblink( $dblink$$x$
        || rsql || $x$$dblink$, false ) AS a( $x$ || v_return_types || ')'
    ;
    PERFORM tools.exec( sql );

    PERFORM tools.db_disconnect();

    -- Add any databases we don't already know about to the databases table
    -- TODO handle race-condition
    INSERT INTO raw.databases( cluster_id, database_oid, database_name )
        SELECT p_cluster.cluster_id, datid, datname
            FROM cluster_databases_t c
            WHERE datname NOT IN ( SELECT database_name FROM raw.databases WHERE cluster_id = p_cluster.cluster_id )
    ;

    -- Create the captured record
    INSERT INTO raw.database_capture( cluster_id, captured_at, previous_capture_id )
        SELECT cluster_id, captured_at, (SELECT capture_id
                                                FROM raw.database_capture c
                                                WHERE c.cluster_id = a.cluster_id
                                                    AND c.captured_at < a.captured_at
                                                ORDER BY captured_at DESC
                                                LIMIT 1 )
            FROM ( SELECT DISTINCT p_cluster.cluster_id AS cluster_id, captured_at FROM cluster_databases_t
                    ) a
    ;

    -- Insert database stats
    INSERT INTO raw.database_stats( database_id, capture_id, num_backends, idle, idle_in_transaction, commits, rollbacks, blocks_read, blocks_hit )
        SELECT database_id, c.capture_id, numbackends, idle, idle_in_transaction, xact_commit, xact_rollback, blks_read, blks_hit
            FROM cluster_databases_t s
                JOIN raw.databases d ON ( d.cluster_id = p_cluster.cluster_id AND d.database_name = s.datname )
                JOIN raw.database_capture c ON ( c.cluster_id = p_cluster.cluster_id AND c.captured_at = s.captured_at )
    ;

    -- Now pull just our databases and call gather.database
    FOR v_database IN SELECT * FROM raw.databases WHERE cluster_id = p_cluster.cluster_id
    LOOP
        PERFORM gather.database( p_cluster, v_database );
    END LOOP;
END;
$gather_cluster$;

CREATE OR REPLACE FUNCTION gather.database(
    p_cluster       public.clusters
    , p_database    raw.databases
) RETURNS void LANGUAGE plpgsql AS $gather_database$
DECLARE
    v_table_name    record;
BEGIN
    -- Connect
    IF NOT tools.db_connect( p_cluster, p_database.database_name ) THEN
        RETURN;
    END IF;

    -- Get all the stat tables
    FOR v_table_name IN SELECT * FROM gather.table_list()
    LOOP
        PERFORM gather.table( p_database.database_id, v_table_name.table_list );
    END LOOP;

    PERFORM tools.db_disconnect();
END;
$gather_database$;

CREATE OR REPLACE FUNCTION gather.table(
    p_database_id   raw.databases.database_id%TYPE
    , p_tablename   text
) RETURNS void LANGUAGE plpgsql AS $gather_table$
DECLARE
    v_return_types  text;
    sql             text;
    rsql            text;
BEGIN
    -- Need to figure out what kind of data is coming back
    v_return_types := tools.table_return_type( p_tablename );

    -- Remote query to run
    rsql := 'SELECT * FROM ' || p_tablename || $$ WHERE schemaname NOT LIKE 'pg_temp_%'$$;

    /*
     * We want to use our clock for recording when data was gathered for
     * consistency, but we don't want to use now() because it could take a
     * non-trivial amount of time to gather all our stats, so use timeofday to
     * get the current time. Because it's getting embedded into the sql at build
     * time, it will still be consistent across the exact dataset that we're
     * gathering.
     */
    sql := 'INSERT INTO ' || p_tablename || $x$_t SELECT $x$ || p_database_id || $x$, '$x$ || timeofday()
            || $x$'::timestamptz, * FROM dblink( $dblink$$x$
            || rsql || $x$$dblink$, false ) AS a( $x$ || v_return_types || ')'
    ;
    PERFORM tools.exec(sql);
END;
$gather_table$;
COMMENT ON FUNCTION gather.table(
    p_database_id   raw.databases.database_id%TYPE
    , p_tablename     text
) IS $$Gathers stats from a single table in a remote database into temp table.

ASSUMPTIONS:
    Temporary tables already exist
    We're already connected to the correct database via the un-named dblink connection

Accepts:
    database_id     ID of the database we're connected to
    tablename       Name of the table to grab
$$;

CREATE OR REPLACE FUNCTION gather.create_table(
    p_tablename     text
) RETURNS void LANGUAGE plpgsql AS $create_table$
DECLARE
    sql text;
BEGIN
    sql := 'CREATE TEMP TABLE ' || p_tablename || '_t AS SELECT * FROM template'
        || CASE WHEN substring(version() from E'***:PostgreSQL (\\d.\\d)') <= '8.2' THEN '_82' ELSE '' END
        || '.' || p_tablename || ' WHERE false'
    ;

    -- We intentionally leave our temp tables behind for debugging (and to conserve OIDS), but that means we need to deal with double-creation
    BEGIN
        PERFORM tools.exec( sql );
    EXCEPTION
        WHEN duplicate_table THEN
            PERFORM tools.exec( 'TRUNCATE ' || p_tablename || '_t' );
    END;
END;
$create_table$;

CREATE OR REPLACE FUNCTION gather.integrate_stats() RETURNS void LANGUAGE plpgsql AS $integrate_stats$
BEGIN
    -- Add any new tables. Ignore any records that only show up in one table, since they belong to an object that was just created or just dropped.
    -- TODO handle race-condition
    INSERT INTO raw.tables( database_id, table_oid, schema_name, table_name )
        SELECT DISTINCT s.database_id, s.relid, s.schemaname, s.relname
            FROM pg_stat_user_tables_t s
                JOIN pg_statio_user_tables_t io USING( database_id, relid )
            WHERE NOT EXISTS( SELECT *
                                FROM raw.tables t
                                WHERE t.database_id = s.database_id
                                    AND t.schema_name = s.schemaname
                                    AND t.table_name = s.relname
                            )
    ;
    
    -- Create capture records.
    INSERT INTO raw.table_capture( database_id, captured_at, previous_capture_id )
        SELECT database_id, captured_at, (SELECT capture_id
                                                FROM raw.table_capture c
                                                WHERE c.database_id = a.database_id
                                                    AND c.captured_at < a.captured_at
                                                ORDER BY captured_at DESC
                                                LIMIT 1 )
            FROM ( SELECT DISTINCT s.database_id, tools.avg( s.captured_at, io.captured_at ) AS captured_at
                        FROM pg_stat_user_tables_t s
                            JOIN pg_statio_user_tables_t io ON ( s.database_id = io.database_id
                                                                    AND s.schemaname = io.schemaname
                                                                    AND s.relname = io.relname
                                                                )
                    ) a
    ;

    -- Now add the data to the stats table
    INSERT INTO raw.table_stats( table_id, capture_id
                , seq_scans, seq_tuples_read, index_scans, index_tuples_fetch, tuples_inserted, tuples_updated, tuples_deleted
                , heap_blocks_read, heap_blocks_hit, index_blocks_read, index_blocks_hit
                , toast_blocks_read, toast_blocks_hit, toast_index_blocks_read, toast_index_blocks_hit
                , last_vacuum, last_autovacuum, last_analyze, last_autoanalyze
        ) SELECT t.table_id, c.capture_id
                , s.seq_scan, s.seq_tup_read, s.idx_scan, s.idx_tup_fetch, s.n_tup_ins, s.n_tup_upd, s.n_tup_del
                , heap_blks_read, heap_blks_hit, idx_blks_read, idx_blks_hit
                , toast_blks_read, toast_blks_hit, tidx_blks_read, tidx_blks_hit
                , last_vacuum, last_autovacuum, last_analyze, last_autoanalyze
            FROM raw.tables t
                JOIN pg_stat_user_tables_t s ON ( t.database_id = s.database_id
                                                    AND t.schema_name = s.schemaname
                                                    AND t.table_name = s.relname
                                                )
                JOIN pg_statio_user_tables_t io ON ( t.database_id = io.database_id
                                                    AND t.schema_name = io.schemaname
                                                    AND t.table_name = io.relname
                                                )
                JOIN raw.table_capture c ON ( c.database_id = t.database_id
                                                AND c.captured_at = tools.avg( s.captured_at, io.captured_at )
                                            )
    ;

    -- For indexes, we cheat a bit and think of them as belonging to a table, even though that's not technically correct
    -- TODO handle race-condition
    INSERT INTO raw.indexes ( database_id, table_id, index_oid, index_name )
        SELECT DISTINCT s.database_id, t.table_id, s.indexrelid, s.indexrelname
            FROM pg_stat_user_indexes_t s
                JOIN pg_statio_user_indexes_t i USING( database_id, indexrelid )
                JOIN raw.tables t ON ( t.database_id = s.database_id
                                        AND t.schema_name = s.schemaname
                                        AND t.table_name = s.relname
                                    )
            WHERE NOT EXISTS( SELECT *
                                FROM raw.indexes i
                                WHERE i.database_id = s.database_id
                                    AND i.table_id = t.table_id
                                    AND i.index_name = s.indexrelname
                            )
    ;
    
    INSERT INTO raw.index_capture( database_id, captured_at, previous_capture_id )
        SELECT database_id, captured_at, (SELECT capture_id
                                                FROM raw.index_capture c
                                                WHERE c.database_id = a.database_id
                                                    AND c.captured_at < a.captured_at
                                                ORDER BY captured_at DESC
                                                LIMIT 1 )
            FROM ( SELECT DISTINCT s.database_id, tools.avg( s.captured_at, io.captured_at ) AS captured_at
                        FROM pg_stat_user_indexes_t s
                            JOIN pg_statio_user_indexes_t io ON ( s.database_id = io.database_id
                                                                    AND s.schemaname = io.schemaname
                                                                    AND s.relname = io.relname
                                                                )
                    ) a
    ;

    INSERT INTO raw.index_stats( index_id, capture_id
                , scans, tuples_read, tuples_fetched
                , blocks_read, blocks_hit
        ) SELECT i.index_id, c.capture_id
                , s.idx_scan, s.idx_tup_read, s.idx_tup_fetch
                , idx_blks_read, idx_blks_hit
            FROM raw.tables t
                JOIN raw.indexes i USING ( database_id, table_id )
                JOIN pg_stat_user_indexes_t s ON ( i.database_id = s.database_id
                                                    AND t.schema_name = s.schemaname
                                                    AND i.index_name = s.indexrelname
                                                )
                JOIN pg_statio_user_indexes_t io ON ( i.database_id = io.database_id
                                                    AND t.schema_name = io.schemaname
                                                    AND i.index_name = io.indexrelname
                                                )
                JOIN raw.index_capture c ON ( c.database_id = t.database_id
                                                AND c.captured_at = tools.avg( s.captured_at, io.captured_at )
                                            )
    ;

    -- Sequences
    -- TODO handle race-condition
    INSERT INTO raw.sequences ( database_id, sequence_oid, schema_name, sequence_name )
        SELECT DISTINCT s.database_id, s.relid, s.schemaname, s.relname
            FROM pg_statio_user_sequences_t s
            WHERE NOT EXISTS( SELECT *
                                FROM raw.sequences t
                                WHERE t.database_id = s.database_id
                                    AND t.schema_name = s.schemaname
                                    AND t.sequence_name = s.relname
                            )
    ;
    
    INSERT INTO raw.sequence_capture( database_id, captured_at, previous_capture_id )
        SELECT database_id, captured_at, (SELECT capture_id
                                                FROM raw.sequence_capture c
                                                WHERE c.database_id = a.database_id
                                                    AND c.captured_at < a.captured_at
                                                ORDER BY captured_at DESC
                                                LIMIT 1 )
            FROM ( SELECT DISTINCT database_id, captured_at
                        FROM pg_statio_user_sequences_t io
                    ) a
    ;


    INSERT INTO raw.sequence_stats( sequence_id, capture_id
                , blocks_read, blocks_hit
        ) SELECT t.sequence_id, c.capture_id
                , blks_read, blks_hit
            FROM raw.sequences t
                JOIN pg_statio_user_sequences_t io ON ( t.database_id = io.database_id
                                                            AND t.schema_name = io.schemaname
                                                            AND t.sequence_name = io.relname
                                                        )
                JOIN raw.sequence_capture c ON ( c.database_id = t.database_id
                                                AND c.captured_at = io.captured_at
                                            )
    ;
END;
$integrate_stats$;

CREATE OR REPLACE FUNCTION tools.exec( sql text ) RETURNS void LANGUAGE plpgsql AS $exec$
BEGIN
    RAISE DEBUG 'Executing: %', sql;
    EXECUTE sql;
END;
$exec$;

CREATE OR REPLACE FUNCTION tools.avg( timestamptz, timestamptz ) RETURNS timestamptz LANGUAGE sql IMMUTABLE AS $avg$
    SELECT ( $2 - $1 ) / 2 + $1
$avg$;

CREATE OR REPLACE FUNCTION tools.db_connect(
    p_cluster               public.clusters
    , p_database_name       raw.databases.database_name%TYPE
) RETURNS boolean LANGUAGE plpgsql AS $db_connect$
BEGIN
    BEGIN
        IF dblink_connect( p_cluster.connect_string || ' dbname=' || p_database_name ) = 'OK' THEN
            RETURN true;
        END IF;
    EXCEPTION
        WHEN sqlclient_unable_to_establish_sqlconnection THEN
            NULL;
    END;
    RAISE WARNING 'Unable to connect to cluster %, database %', p_cluster.cluster_name, p_database_name;
    RETURN false;
END;
$db_connect$;
CREATE OR REPLACE FUNCTION tools.db_disconnect(
) RETURNS boolean LANGUAGE plpgsql AS $db_disconnect$
BEGIN
    IF dblink_disconnect() = 'OK' THEN
        RETURN true;
    END IF;
    RAISE WARNING 'Unable to disconnect';
    RETURN false;
END;
$db_disconnect$;

CREATE OR REPLACE FUNCTION tools.table_return_type(
    p_tablename     text
) RETURNS text LANGUAGE plpgsql AS $table_return_data$
DECLARE
    sql text;
    result text;
BEGIN
    sql := $$SELECT a.attname::text, pg_catalog.format_type(a.atttypid, a.atttypmod)
                FROM pg_catalog.pg_attribute a
                WHERE a.attrelid = '$$ || p_tablename || $$'::regclass AND a.attnum > 0 AND NOT a.attisdropped
                ORDER BY a.attnum$$;
    
    RAISE DEBUG 'Remote exec on nameless connection: %', sql;
    SELECT INTO result array_to_string( array(
            SELECT attname || ' ' || CASE WHEN att_type = 'oid' THEN 'bigint' ELSE att_type END
                FROM dblink( sql, true ) AS a( attname text, att_type text )
        ), ', ' )
    ;
    RAISE DEBUG 'return types: %', result;

    RETURN result;
END;
$table_return_data$;

CREATE OR REPLACE FUNCTION delta.add(
    p_max_captures      int
) RETURNS int LANGUAGE plpgsql AS $add$
DECLARE
    i int := 0;
    v_rows int := 0;
BEGIN
    /*
        All of these queries operate the same, just on different tables.

        First, we get the set of distinct database_id's (or table, index, etc) and captured_on times.
        This gives us every capture run that's occured. As part of that subquery, we exclude any captured_on
        times that are less than the maximum captured_on time in the delta table. This ensures we're not 
        putting any duplicate values in, but it also means that we won't add any data that was added after
        processing. This isn't a concern as long as the system clock on the machine collecting stats never
        runs backwards. We also exclude captured_on times that exceed a specified interval, as a way to
        throttle how much work we'll do at once.

        Next, we add to that set the previous capture time for each {object_id, captured_at} tuple. For each
        object_id, we now have the set of captured_at times we're interested in, as well as the previous
        capture time for each one.

        Once we have the captured_at and previous data, we use that to join to the object table twice: once
        to get the captured_at record, and once to get the previous record. After that, it's just a matter
        of math to come up with the deltas between each capture time.
    */

    CREATE TEMP TABLE database_capture AS
        SELECT DISTINCT database_id, captured_at
            FROM raw.database_stats r
            WHERE captured_at > coalesce( (SELECT max(captured_at)
                                                FROM delta.database_stats d
                                                WHERE d.database_id = r.database_id)
                                            , '-infinity'::timestamptz )
    ;
    CREATE INDEX database_capture__captured ON database_capture( captured_at );
    INSERT INTO delta.database_stats( database_id, captured_at, previous_time
            , num_backends, idle, idle_in_transaction
            , commits, rollbacks, blocks_read, blocks_hit )
        SELECT d.database_id, d.captured_at, d.previous AS previous_time
                , c.num_backends, c.idle, c.idle_in_transaction
                , delta.delta( p.commits, c.commits) AS commits
                , delta.delta( p.rollbacks, c.rollbacks) AS rollbacks
                , delta.delta( p.blocks_read, c.blocks_read) AS blocks_read
                , delta.delta( p.blocks_hit, c.blocks_hit) AS blocks_hit
        FROM raw.database_stats c
            JOIN raw.database_stats p USING ( database_id )
            -- Get the set of all the captured_at times for each database, as well as the previous captured_at time for each one
            JOIN ( SELECT database_id, captured_at
                            , (SELECT max(captured_at)
                                    FROM raw.database_stats c
                                    WHERE c.database_id = capture.database_id
                                        AND c.captured_at < capture.captured_at
                                ) AS previous
                        -- Get all the captured_at times for each database that are newer than the latest record in the delta database
                        FROM database_capture capture
                        WHERE captured_at <= (SELECT max(captured_at)
                                                FROM (SELECT DISTINCT captured_at
                                                        FROM database_capture
                                                        ORDER BY captured_at
                                                        LIMIT p_max_captures
                                                        ) a
                                                ) 
                    ) d ON ( c.database_id = d.database_id AND c.captured_at = d.captured_at AND p.captured_at = d.previous )
    ;
    GET DIAGNOSTICS i = ROW_COUNT;
    v_rows := v_rows + i;

    CREATE TEMP TABLE table_capture AS
        SELECT DISTINCT table_id, captured_at
            FROM raw.table_stats r
            WHERE captured_at > coalesce( (SELECT max(captured_at)
                                                FROM delta.table_stats d
                                                WHERE d.table_id = r.table_id)
                                            , '-infinity'::timestamptz )
    ;
    CREATE INDEX table_capture__captured ON table_capture( captured_at );
    INSERT INTO delta.table_stats ( table_id, captured_at, previous_time
            , seq_scans, seq_tuples_read, index_scans, index_tuples_fetch
            , tuples_inserted , tuples_updated , tuples_deleted
            , heap_blocks_read, heap_blocks_hit , index_blocks_read, index_blocks_hit
            , toast_blocks_read, toast_blocks_hit , toast_index_blocks_read, toast_index_blocks_hit
            , last_vacuum, last_autovacuum , last_analyze, last_autoanalyze )
        SELECT d.table_id, d.captured_at, d.previous AS previous_time
                , delta.delta( p.seq_scans, c.seq_scans )
                , delta.delta( p.seq_tuples_read, c.seq_tuples_read )
                , delta.delta( p.index_scans, c.index_scans )
                , delta.delta( p.index_tuples_fetch, c.index_tuples_fetch )
                , delta.delta( p.tuples_inserted, c.tuples_inserted )
                , delta.delta( p.tuples_updated, c.tuples_updated )
                , delta.delta( p.tuples_deleted, c.tuples_deleted )
                , delta.delta( p.heap_blocks_read, c.heap_blocks_read )
                , delta.delta( p.heap_blocks_hit, c.heap_blocks_hit )
                , delta.delta( p.index_blocks_read, c.index_blocks_read )
                , delta.delta( p.index_blocks_hit, c.index_blocks_hit )
                , delta.delta( p.toast_blocks_read, c.toast_blocks_read )
                , delta.delta( p.toast_blocks_hit, c.toast_blocks_hit )
                , delta.delta( p.toast_index_blocks_read, c.toast_index_blocks_read )
                , delta.delta( p.toast_index_blocks_hit, c.toast_index_blocks_hit )
                , c.last_vacuum, c.last_autovacuum, c.last_analyze, c.last_autoanalyze
        FROM raw.table_stats c
            JOIN raw.table_stats p USING ( table_id )
            JOIN ( SELECT table_id, captured_at
                            , (SELECT max(captured_at)
                                    FROM raw.table_stats c
                                    WHERE c.table_id = capture.table_id
                                        AND c.captured_at < capture.captured_at
                                ) AS previous
                        FROM table_capture capture
                        WHERE captured_at <= (SELECT max(captured_at)
                                                FROM (SELECT DISTINCT captured_at
                                                        FROM table_capture
                                                        ORDER BY captured_at
                                                        LIMIT p_max_captures
                                                        ) a
                                                ) 
                    ) d ON ( c.table_id = d.table_id AND c.captured_at = d.captured_at AND p.captured_at = d.previous )
    ;
    GET DIAGNOSTICS i = ROW_COUNT;
    v_rows := v_rows + i;

    CREATE TEMP TABLE index_capture AS
        SELECT DISTINCT index_id, captured_at
            FROM raw.index_stats r
            WHERE captured_at > coalesce( (SELECT max(captured_at)
                                                FROM delta.index_stats d
                                                WHERE d.index_id = r.index_id)
                                            , '-infinity'::timestamptz )
    ;
    CREATE INDEX index_capture__captured ON index_capture( captured_at );
    INSERT INTO delta.index_stats ( index_id, captured_at, previous_time
            , scans, tuples_read, tuples_fetched
            , blocks_read, blocks_hit )
        SELECT d.index_id, d.captured_at, d.previous AS previous_time
                , delta.delta( p.scans, c.scans )
                , delta.delta( p.tuples_read, c.tuples_read )
                , delta.delta( p.tuples_fetched, c.tuples_fetched )
                , delta.delta( p.blocks_read, c.blocks_read )
                , delta.delta( p.blocks_hit, c.blocks_hit )
        FROM raw.index_stats c
            JOIN raw.index_stats p USING ( index_id )
            JOIN ( SELECT index_id, captured_at
                            , (SELECT max(captured_at)
                                    FROM raw.index_stats c
                                    WHERE c.index_id = capture.index_id
                                        AND c.captured_at < capture.captured_at
                                ) AS previous
                        FROM index_capture capture
                        WHERE captured_at <= (SELECT max(captured_at)
                                                FROM (SELECT DISTINCT captured_at
                                                        FROM index_capture
                                                        ORDER BY captured_at
                                                        LIMIT p_max_captures
                                                        ) a
                                                ) 
                    ) d ON ( c.index_id = d.index_id AND c.captured_at = d.captured_at AND p.captured_at = d.previous )
    ;
    GET DIAGNOSTICS i = ROW_COUNT;
    v_rows := v_rows + i;

    CREATE TEMP TABLE sequence_capture AS
        SELECT DISTINCT sequence_id, captured_at
            FROM raw.sequence_stats r
            WHERE captured_at > coalesce( (SELECT max(captured_at)
                                                FROM delta.sequence_stats d
                                                WHERE d.sequence_id = r.sequence_id)
                                            , '-infinity'::timestamptz )
    ;
    CREATE INDEX sequence_capture__captured ON sequence_capture( captured_at );
    INSERT INTO delta.sequence_stats ( sequence_id, captured_at, previous_time
            , blocks_read, blocks_hit )
        SELECT d.sequence_id, d.captured_at, d.previous AS previous_time
                , delta.delta( p.blocks_read, c.blocks_read )
                , delta.delta( p.blocks_hit, c.blocks_hit )
        FROM raw.sequence_stats c
            JOIN raw.sequence_stats p USING ( sequence_id )
            JOIN ( SELECT sequence_id, captured_at
                            , (SELECT max(captured_at)
                                    FROM raw.sequence_stats c
                                    WHERE c.sequence_id = capture.sequence_id
                                        AND c.captured_at < capture.captured_at
                                ) AS previous
                        FROM sequence_capture capture
                        WHERE captured_at <= (SELECT max(captured_at)
                                                FROM (SELECT DISTINCT captured_at
                                                        FROM sequence_capture
                                                        ORDER BY captured_at
                                                        LIMIT p_max_captures
                                                        ) a
                                                ) 
                    ) d ON ( c.sequence_id = d.sequence_id AND c.captured_at = d.captured_at AND p.captured_at = d.previous )
    ;
    GET DIAGNOSTICS i = ROW_COUNT;
    v_rows := v_rows + i;

    RETURN v_rows;
END;
$add$;
CREATE OR REPLACE FUNCTION delta.add() RETURNS int LANGUAGE sql AS $add$
SELECT delta.add( 2000000000 )
$add$;

CREATE OR REPLACE FUNCTION delta.delta(
    p_first     bigint
    , p_second  bigint
) RETURNS bigint LANGUAGE sql AS $delta$
    SELECT CASE WHEN $2 >= $1 THEN
            -- Normal case
            $2 - $1
        WHEN $1 > 9223372036854775807 - 1000000000 THEN
            9223372036854775807 - $1 + $2
        ELSE NULL
        END
$delta$;
COMMENT ON FUNCTION delta.delta(bigint, bigint) IS $$Calculates difference between two data points, accounting for rollover.
To protect against arbitrary resets, the rollover code is only active within max_value - 1B.$$;

CREATE OR REPLACE FUNCTION delta.per_second(
    p_first     timestamptz
    , p_second  timestamptz
    , p_delta     bigint
) RETURNS float LANGUAGE sql AS $per_second$
    SELECT $3 / extract( epoch from $2 - $1 )
$per_second$;
COMMENT ON FUNCTION delta.per_second(timestamptz, timestamptz, bigint) IS $$Converts p_delta into a per-second value$$;

--		vi: expandtab sw=4 ts=4
