/*
 * Copyright (c) 2007, Jim C. Nasby (decibel@decibel.org)
 * All rights reserved.
 *
 * $Id: schema.sql 37 2008-11-05 20:19:20Z decibel $
 */

CREATE SCHEMA gather;
CREATE SCHEMA tools;
CREATE SCHEMA raw;
CREATE SCHEMA delta;

CREATE TABLE public.clusters (
	cluster_id				serial	PRIMARY KEY
	, cluster_name			text	NOT NULL UNIQUE
	, connect_string		text	NOT NULL
	, exclude_databases		text[]
);

CREATE TABLE raw.databases (
	database_id			serial	PRIMARY KEY
	, cluster_id		int		NOT NULL REFERENCES public.clusters( cluster_id ) ON DELETE CASCADE
	, database_oid		bigint	NOT NULL
	, database_name		text	NOT NULL
	, UNIQUE( cluster_id, database_name )
);

CREATE TABLE raw.database_capture (
	capture_id				serial	PRIMARY KEY
	, cluster_id			int		NOT NULL REFERENCES public.clusters( cluster_id ) ON DELETE CASCADE
	, captured_at			timestamptz NOT NULL
	, previous_capture_id	int
	, UNIQUE( cluster_id, captured_at )
);
CREATE TABLE raw.database_stats (
	database_id				int	NOT NULL REFERENCES raw.databases( database_id ) ON DELETE CASCADE
	, capture_id			int	NOT NULL REFERENCES raw.database_capture( capture_id ) ON DELETE CASCADE
	, num_backends			integer
	, idle					int
	, idle_in_transaction	int
	, commits			bigint
	, rollbacks			bigint
	, blocks_read			bigint
	, blocks_hit			bigint
	, PRIMARY KEY( database_id, capture_id )
);
CREATE INDEX database_stats__capture_id ON raw.database_stats( capture_id );

CREATE TABLE raw.tables (
	table_id		serial	PRIMARY KEY
	, database_id	int		NOT NULL REFERENCES raw.databases( database_id ) ON DELETE CASCADE
	, table_oid		bigint	NOT NULL
	, schema_name	text	NOT NULL
	, table_name	text	NOT NULL
	, UNIQUE( database_id, schema_name, table_name )
);

CREATE TABLE raw.table_capture (
	capture_id				serial	PRIMARY KEY
	, database_id			int		NOT NULL REFERENCES raw.databases( database_id ) ON DELETE CASCADE
	, captured_at			timestamptz NOT NULL
	, previous_capture_id	int
	, UNIQUE( database_id, captured_at )
);
CREATE TABLE raw.table_stats (
	table_id					int			NOT NULL REFERENCES raw.tables(table_id) ON DELETE CASCADE
	, capture_id				int			NOT NULL REFERENCES raw.table_capture( capture_id ) ON DELETE CASCADE
	, seq_scans					bigint
	, seq_tuples_read			bigint
	, index_scans				bigint
	, index_tuples_fetch		bigint
	, tuples_inserted			bigint
	, tuples_updated			bigint
	, tuples_deleted			bigint
	, heap_blocks_read			bigint
	, heap_blocks_hit			bigint
	, index_blocks_read			bigint
	, index_blocks_hit			bigint
	, toast_blocks_read			bigint
	, toast_blocks_hit			bigint
	, toast_index_blocks_read	bigint
	, toast_index_blocks_hit	bigint
	, last_vacuum				timestamp with time zone
	, last_autovacuum			timestamp with time zone
	, last_analyze				timestamp with time zone
	, last_autoanalyze			timestamp with time zone
	, PRIMARY KEY( table_id, capture_id )
);
CREATE INDEX table_stats__capture_id ON raw.table_stats( capture_id );

CREATE TABLE raw.indexes (
	index_id		serial	PRIMARY KEY
	, database_id	int		NOT NULL REFERENCES raw.databases( database_id ) ON DELETE CASCADE
	, table_id		int		NOT NULL REFERENCES raw.tables( table_id ) ON DELETE CASCADE
	, index_oid		bigint	NOT NULL
	, index_name	text	NOT NULL
	, UNIQUE( database_id, table_id, index_name )
);

CREATE TABLE raw.index_capture (
	capture_id				serial	PRIMARY KEY
	, database_id			int		NOT NULL REFERENCES raw.databases( database_id ) ON DELETE CASCADE
	, captured_at			timestamptz NOT NULL
	, previous_capture_id	int
	, UNIQUE( database_id, captured_at )
);
CREATE TABLE raw.index_stats (
	index_id			int			NOT NULL REFERENCES raw.indexes(index_id) ON DELETE CASCADE
	, capture_id		int			NOT NULL REFERENCES raw.index_capture( capture_id ) ON DELETE CASCADE
	, scans				bigint
	, tuples_read		bigint
	, tuples_fetched	bigint
	, blocks_read		bigint
	, blocks_hit		bigint
	, PRIMARY KEY( index_id, capture_id )
);
CREATE INDEX index_stats__capture_id ON raw.index_stats( capture_id );

CREATE TABLE raw.sequences (
	sequence_id		serial	PRIMARY KEY
	, database_id	int		NOT NULL REFERENCES raw.databases( database_id ) ON DELETE CASCADE
	, sequence_oid	bigint	NOT NULL
	, schema_name	text	NOT NULL
	, sequence_name	text	NOT NULL
	, UNIQUE( database_id, schema_name, sequence_name )
);

CREATE TABLE raw.sequence_capture (
	capture_id				serial	PRIMARY KEY
	, database_id			int		NOT NULL REFERENCES raw.databases( database_id ) ON DELETE CASCADE
	, captured_at			timestamptz NOT NULL
	, previous_capture_id	int
	, UNIQUE( database_id, captured_at )
);
CREATE TABLE raw.sequence_stats (
	sequence_id			int			NOT NULL REFERENCES raw.sequences(sequence_id) ON DELETE CASCADE
	, capture_id		int			NOT NULL REFERENCES raw.sequence_capture( capture_id ) ON DELETE CASCADE
	, blocks_read		bigint
	, blocks_hit		bigint
	, PRIMARY KEY( sequence_id, capture_id )
);
CREATE INDEX sequence_stats__capture_id ON raw.sequence_stats( capture_id );
